# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import io
from typing import Callable, List, Optional, Sequence, Union

import imageio
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from gpflow.models import GPModel
from matplotlib import cm
from matplotlib.axes import Axes
from matplotlib.collections import Collection
from matplotlib.colors import rgb2hex
from matplotlib.contour import ContourSet
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle

from trieste.acquisition import AcquisitionFunction
from trieste.acquisition.multi_objective.dominance import non_dominated
from trieste.bayesian_optimizer import FrozenRecord, Record, StateType
from trieste.models import ProbabilisticModel
from trieste.observer import OBJECTIVE
from trieste.space import TaggedMultiSearchSpace
from trieste.types import TensorType
from trieste.utils import to_numpy
from trieste.utils.misc import LocalizedTag


def create_grid(
    mins: TensorType, maxs: TensorType, grid_density: int = 30
) -> tuple[TensorType, TensorType, TensorType]:
    """
    Creates a regular 2D grid of size `grid_density^2` between mins and maxs.

    :param mins: list of 2 lower bounds
    :param maxs: list of 2 upper bounds
    :param grid_density: scalar
    :return: Xplot [grid_density**2, 2], xx, yy from meshgrid for the specific formatting of
        contour / surface plots
    """
    tf.debugging.assert_shapes([(mins, [2]), (maxs, [2])])

    xspaced = np.linspace(mins[0], maxs[0], grid_density)
    yspaced = np.linspace(mins[1], maxs[1], grid_density)
    xx, yy = np.meshgrid(xspaced, yspaced)
    Xplot = np.vstack((xx.flatten(), yy.flatten())).T

    return Xplot, xx, yy


def plot_surface(
    xx: TensorType,
    yy: TensorType,
    f: TensorType,
    ax: Axes,
    contour: bool = False,
    fill: bool = False,
    alpha: float = 1.0,
) -> ContourSet | Collection:
    """
    Adds either a contour or a surface to a given ax.

    :param xx: input 1, from meshgrid
    :param yy: input2, from meshgrid
    :param f: output, from meshgrid
    :param ax: plt axes object
    :param contour: Boolean
    :param fill: filled contour
    :param alpha: transparency
    :return: generated contour or surface
    """

    if contour:
        if fill:
            return ax.contourf(xx, yy, f.reshape(*xx.shape), 80, alpha=alpha)
        else:
            return ax.contour(xx, yy, f.reshape(*xx.shape), 80, alpha=alpha)
    else:
        return ax.plot_surface(
            xx,
            yy,
            f.reshape(*xx.shape),
            cmap=cm.coolwarm,
            linewidth=0,
            antialiased=False,
            alpha=alpha,
        )


def plot_function_2d(
    obj_func: Callable[[TensorType], TensorType],
    mins: TensorType,
    maxs: TensorType,
    grid_density: int = 100,
    contour: bool = False,
    log: bool = False,
    title: Optional[Sequence[str]] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    figsize: Optional[tuple[float, float]] = (8, 6),
    colorbar: bool = False,
    alpha: float = 1.0,
    fill: bool = False,
) -> tuple[Figure, Axes]:
    """
    2D/3D plot of an obj_func for a grid of size grid_density**2 between mins and maxs

    :param obj_func: a function that returns a n-array given a [n, d] array
    :param mins: 2 lower bounds
    :param maxs: 2 upper bounds
    :param grid_density: positive integer for the grid size
    :param contour: Boolean. If False, a 3d plot is produced
    :param log: Boolean. If True, the log transformation (log(f - min(f) + 0.1)) is applied
    :param title: optional titles
    :param xlabel: optional xlabel
    :param ylabel: optional ylabel
    :param figsize: optional figsize
    :param colorbar: whether to use colorbar
    :param alpha: transparency
    :param fill: filled contour
    :return: figure and axes
    """
    mins = to_numpy(mins)
    maxs = to_numpy(maxs)

    # Create a regular grid on the parameter space
    Xplot, xx, yy = create_grid(mins=mins, maxs=maxs, grid_density=grid_density)

    # Evaluate objective function
    F = to_numpy(obj_func(Xplot))
    if len(F.shape) == 1:
        F = F.reshape(-1, 1)

    n_output = F.shape[1]

    if contour:
        fig, ax = plt.subplots(
            1, n_output, squeeze=False, sharex="all", sharey="all", figsize=figsize
        )
    else:
        fig = plt.figure(figsize=figsize)

    for k in range(F.shape[1]):
        # Apply log transformation
        f = F[:, k]
        if log:
            f = np.log(f - np.min(f) + 1e-1)

        # Either plot contour of surface
        if contour:
            axx = ax[0, k]
        else:
            ax = axx = fig.add_subplot(1, n_output, k + 1, projection="3d")

        plt_obj = plot_surface(xx, yy, f, axx, contour=contour, alpha=alpha, fill=fill)
        if title is not None:
            axx.set_title(title[k])
        if colorbar:
            fig.colorbar(plt_obj, ax=axx)
        axx.set_xlabel(xlabel)
        axx.set_ylabel(ylabel)
        axx.set_xlim(mins[0], maxs[0])
        axx.set_ylim(mins[1], maxs[1])

    return fig, ax


def plot_acq_function_2d(
    acq_func: AcquisitionFunction,
    mins: TensorType,
    maxs: TensorType,
    grid_density: int = 100,
    contour: bool = False,
    log: bool = False,
    title: Optional[Sequence[str]] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    figsize: Optional[tuple[float, float]] = (8, 6),
    colorbar: bool = False,
    alpha: float = 1.0,
    fill: bool = False,
) -> tuple[Figure, Axes]:
    """
    Wrapper to produce a 2D/3D plot of an acq_func for a grid of size grid_density**2 between
    mins and maxs.

    :param acq_func: a function that returns a n-array given a [n, d] array
    :param mins: 2 lower bounds
    :param maxs: 2 upper bounds
    :param grid_density: positive integer for the grid size
    :param contour: Boolean. If False, a 3d plot is produced
    :param log: Boolean. If True, the log transformation (log(f - min(f) + 0.1)) is applied
    :param title: optional titles
    :param xlabel: optional xlabel
    :param ylabel: optional ylabel
    :param figsize: optional figsize
    :param colorbar: whether to use colorbar
    :param alpha: transparency
    :param fill: filled contour
    :return: figure and axes
    """

    def batched_func(x: TensorType) -> TensorType:
        return acq_func(tf.expand_dims(x, axis=-2))

    return plot_function_2d(
        batched_func,
        mins,
        maxs,
        grid_density,
        contour,
        log,
        title,
        xlabel,
        ylabel,
        figsize,
        colorbar,
        alpha,
        fill,
    )


def format_point_markers(
    num_pts: int,
    num_init: Optional[Union[int, TensorType]] = None,
    idx_best: Optional[TensorType] = None,
    mask_fail: Optional[TensorType] = None,
    m_init: str = "x",
    m_add: str = "o",
    c_pass: str = "tab:green",
    c_fail: Union[str, List[str]] = "tab:red",
    c_best: str = "tab:purple",
) -> tuple[TensorType, TensorType]:
    """
    Prepares point marker styles according to some BO factors.

    :param num_pts: total number of BO points
    :param num_init: initial number of BO points; can also be a mask
    :param idx_best: index of the best BO point(s)
    :param mask_fail: Bool vector, True if the corresponding observation violates the constraint(s)
    :param m_init: marker for the initial BO points
    :param m_add: marker for the other BO points
    :param c_pass: color for the regular BO points
    :param c_fail: color for the failed BO points
    :param c_best: color for the best BO points
    :return: 2 string vectors col_pts, mark_pts containing marker styles and colors
    """
    if num_init is None:
        num_init = num_pts
    col_pts = np.repeat(c_pass, num_pts)
    col_pts = col_pts.astype("<U15")
    mark_pts = np.repeat(m_init, num_pts)
    if isinstance(num_init, int):
        mark_pts[num_init:] = m_add
    else:
        mark_pts[np.where(~num_init)] = m_add
    if mask_fail is not None:
        col_pts[np.where(mask_fail)] = c_fail
    if idx_best is not None:
        col_pts[idx_best] = c_best

    return col_pts, mark_pts


def plot_bo_points(
    pts: TensorType,
    ax: Axes,
    num_init: Optional[Union[int, TensorType]] = None,
    idx_best: Optional[int] = None,
    mask_fail: Optional[TensorType] = None,
    obs_values: Optional[TensorType] = None,
    m_init: str = "x",
    m_add: str = "o",
    c_pass: str = "tab:green",
    c_fail: Union[str, List[str]] = "tab:red",
    c_best: str = "tab:purple",
) -> None:
    """
    Adds scatter points to an existing subfigure. Markers and colors are chosen according to
    BO factors.

    :param pts: [N, 2] x inputs
    :param ax: a plt axes object
    :param num_init: initial number of BO points; can also be a mask
    :param idx_best: index of the best BO point
    :param mask_fail: Bool vector, True if the corresponding observation violates the constraint(s)
    :param obs_values: optional [N] outputs (for 3d plots)
    :param m_init: marker for the initial BO points
    :param m_add: marker for the other BO points
    :param c_pass: color for the regular BO points
    :param c_fail: color for the failed BO points
    :param c_best: color for the best BO points
    """

    num_pts = pts.shape[0]

    col_pts, mark_pts = format_point_markers(
        num_pts, num_init, idx_best, mask_fail, m_init, m_add, c_pass, c_fail, c_best
    )

    if obs_values is None:
        for i in range(pts.shape[0]):
            ax.scatter(pts[i, 0], pts[i, 1], c=col_pts[i], marker=mark_pts[i])
    else:
        for i in range(pts.shape[0]):
            ax.scatter(pts[i, 0], pts[i, 1], obs_values[i], c=col_pts[i], marker=mark_pts[i])


def plot_mobo_points_in_obj_space(
    obs_values: TensorType,
    num_init: Optional[int] = None,
    mask_fail: Optional[TensorType] = None,
    figsize: Optional[tuple[float, float]] = (8, 6),
    xlabel: str = "Obj 1",
    ylabel: str = "Obj 2",
    zlabel: str = "Obj 3",
    title: Optional[str] = None,
    m_init: str = "x",
    m_add: str = "o",
    c_pass: str = "tab:green",
    c_fail: str = "tab:red",
    c_pareto: str = "tab:purple",
    only_plot_pareto: bool = False,
) -> tuple[Figure, Axes]:
    """
    Adds scatter points in objective space, used for multi-objective optimization (2 or 3
    objectives only). Markers and colors are chosen according to BO factors.

    :param obs_values: TF Tensor or numpy array of objective values, shape (N, 2) or (N, 3).
    :param num_init: initial number of BO points
    :param mask_fail: Bool vector, True if the corresponding observation violates the constraint(s)
    :param figsize: Size of the figure.
    :param xlabel: Label of the X axis.
    :param ylabel: Label of the Y axis.
    :param zlabel: Label of the Z axis (in 3d case).
    :param title: Title of the plot.
    :param m_init: Marker for initial points.
    :param m_add: Marker for the points observed during the BO loop.
    :param c_pass: color for the regular BO points
    :param c_fail: color for the failed BO points
    :param c_pareto: color for the Pareto front points
    :param only_plot_pareto: if set to `True`, only plot the pareto points. Default is `False`.
    :return: figure and axes
    """
    obj_num = obs_values.shape[-1]
    tf.debugging.assert_shapes([])
    assert obj_num in (2, 3), NotImplementedError(
        f"Only support 2/3-objective functions but found: {obj_num}"
    )

    _, dom = non_dominated(obs_values)
    idx_pareto = np.where(dom) if mask_fail is None else np.where(np.logical_and(dom, ~mask_fail))

    pts = obs_values.numpy() if tf.is_tensor(obs_values) else obs_values
    num_pts = pts.shape[0]

    col_pts, mark_pts = format_point_markers(
        num_pts, num_init, idx_pareto, mask_fail, m_init, m_add, c_pass, c_fail, c_pareto
    )
    if only_plot_pareto:
        col_pts = col_pts[idx_pareto]
        mark_pts = mark_pts[idx_pareto]
        pts = pts[idx_pareto]

    if obj_num == 2:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
    for i in range(pts.shape[0]):
        ax.scatter(*pts[i], c=col_pts[i], marker=mark_pts[i])
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if obj_num == 3:
        ax.set_zlabel(zlabel)
    if title is not None:
        ax.set_title(title)
    return fig, ax


def plot_mobo_history(
    obs_values: TensorType,
    metric_func: Callable[[TensorType], TensorType],
    num_init: int,
    mask_fail: Optional[TensorType] = None,
    figsize: Optional[tuple[float, float]] = (8, 6),
) -> tuple[Figure, Axes]:
    """
    Draw the performance measure for multi-objective optimization.

    :param obs_values: TF Tensor or numpy array of objective values
    :param metric_func: a callable function calculate metric score
    :param num_init: initial number of BO points
    :param mask_fail: Bool vector, True if the corresponding observation violates the constraint(s)
    :param figsize: Size of the figure.
    :return: figure and axes
    """

    fig, ax = plt.subplots(figsize=figsize)
    size, obj_num = obs_values.shape

    if mask_fail is not None:
        obs_values[mask_fail] = [np.inf] * obj_num

    _idxs = np.arange(1, size + 1)
    ax.plot(_idxs, [metric_func(obs_values[:pts, :]) for pts in _idxs], color="tab:orange")
    ax.axvline(x=num_init - 0.5, color="tab:blue")
    return fig, ax


def plot_regret(
    obs_values: TensorType,
    ax: Axes,
    num_init: int,
    show_obs: bool = True,
    mask_fail: Optional[TensorType] = None,
    idx_best: Optional[int] = None,
    m_init: str = "x",
    m_add: str = "o",
    c_pass: str = "tab:green",
    c_fail: str = "tab:red",
    c_best: str = "tab:purple",
) -> None:
    """
    Draws the simple regret with same colors / markers as the other plots.

    :param obs_values: TF Tensor or numpy array of objective values
    :param ax: a plt axes object
    :param show_obs: show observations
    :param num_init: initial number of BO points
    :param mask_fail: Bool vector, True if the corresponding observation violates the constraint(s)
    :param idx_best: index of the best BO point
    :param m_init: marker for the initial BO points
    :param m_add: marker for the other BO points
    :param c_pass: color for the regular BO points
    :param c_fail: color for the failed BO points
    :param c_best: color for the best BO points
    """

    col_pts, mark_pts = format_point_markers(
        obs_values.shape[0], num_init, idx_best, mask_fail, m_init, m_add, c_pass, c_fail, c_best
    )

    safe_obs_values = obs_values.copy()
    if mask_fail is not None:
        safe_obs_values[mask_fail] = np.max(obs_values)

    ax.plot(np.minimum.accumulate(safe_obs_values), color="tab:orange")

    if show_obs:
        for i in range(obs_values.shape[0]):
            ax.scatter(i, obs_values[i], c=col_pts[i], marker=mark_pts[i])

    ax.axvline(x=num_init - 0.5, color="tab:blue")


def plot_gp_2d(
    model: GPModel,
    mins: TensorType,
    maxs: TensorType,
    grid_density: int = 100,
    contour: bool = False,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    figsize: Optional[tuple[float, float]] = (8, 6),
    predict_y: bool = False,
) -> tuple[Figure, Axes]:
    """
    2D/3D plot of a gp model for a grid of size grid_density**2 between mins and maxs

    :param model: a gpflow model
    :param mins: 2 lower bounds
    :param maxs: 2 upper bounds
    :param grid_density: positive integer for the grid size
    :param contour: Boolean. If False, a 3d plot is produced
    :param xlabel: optional string
    :param ylabel: optional string
    :param figsize: optional figsize
    :param predict_y: predict_y or predict_f
    :return: figure and axes
    """
    mins = to_numpy(mins)
    maxs = to_numpy(maxs)

    # Create a regular grid on the parameter space
    Xplot, xx, yy = create_grid(mins=mins, maxs=maxs, grid_density=grid_density)

    # Evaluate objective function
    if predict_y:
        Fmean, Fvar = model.predict_y(Xplot)
    else:
        Fmean, Fvar = model.predict_f(Xplot)

    n_output = Fmean.shape[1]

    if contour:
        fig, ax = plt.subplots(
            n_output, 2, squeeze=False, sharex="all", sharey="all", figsize=figsize
        )
        ax[0, 0].set_xlim(mins[0], maxs[0])
        ax[0, 0].set_ylim(mins[1], maxs[1])
    else:
        fig = plt.figure(figsize=figsize)

    for k in range(n_output):
        # Apply log transformation
        fmean = Fmean[:, k].numpy()
        fvar = Fvar[:, k].numpy()

        # Either plot contour of surface
        if contour:
            axx = ax[k, 0]
            plot_surface(xx, yy, fmean, ax[k, 0], contour=contour, alpha=1.0)
            plot_surface(xx, yy, fvar, ax[k, 1], contour=contour, alpha=1.0)
            ax[k, 0].set_title("mean")
            ax[k, 1].set_title("variance")
            ax[k, 0].set_xlabel(xlabel)
            ax[k, 0].set_ylabel(ylabel)
            ax[k, 1].set_xlabel(xlabel)
            ax[k, 1].set_ylabel(ylabel)
        else:
            ax = axx = fig.add_subplot(1, n_output, k + 1, projection="3d")
            plot_surface(xx, yy, fmean, axx, contour=contour, alpha=0.5)
            ucb = fmean + 2.0 * np.sqrt(fvar)
            lcb = fmean - 2.0 * np.sqrt(fvar)
            plot_surface(xx, yy, ucb, axx, contour=contour, alpha=0.1)
            plot_surface(xx, yy, lcb, axx, contour=contour, alpha=0.1)
            axx.set_xlabel(xlabel)
            axx.set_ylabel(ylabel)
            axx.set_xlim(mins[0], maxs[0])
            axx.set_ylim(mins[1], maxs[1])

    return fig, ax


def plot_trust_region_history_2d(
    obj_func: Callable[[TensorType], TensorType],
    mins: TensorType,
    maxs: TensorType,
    history: Record[StateType, ProbabilisticModel] | FrozenRecord[StateType, ProbabilisticModel],
    num_query_points: Optional[int] = None,
    num_init: Optional[int] = None,
    alpha: float = 0.3,
) -> tuple[Optional[Figure], Optional[Axes]]:
    """
    Plot the contour of the objective function, query points and the trust regions for a particular
    step of the optimization process.

    :param obj_func: the objective function that returns a n-array given a [n, d] array
    :param mins: search space 2D lower bounds
    :param maxs: search space 2D upper bounds
    :param history: the optimization history for a particular step of the optimization process
    :param num_query_points: total number of query points in this step
    :param num_init: initial number of BO points
    :param alpha: transparency for the trust regions
    :return: figure and axes
    """

    state = history.acquisition_state
    if state is None:
        # If state is None, then there is no trust region state to plot.
        return None, None

    # Plot objective contour.
    fig, ax = plot_function_2d(
        obj_func,
        mins,
        maxs,
        grid_density=40,
        contour=True,
    )

    assert hasattr(state, "acquisition_space")
    acquisition_space = state.acquisition_space

    if isinstance(acquisition_space, TaggedMultiSearchSpace):
        spaces = [acquisition_space.get_subspace(tag) for tag in acquisition_space.subspace_tags]
    else:
        spaces = [acquisition_space]

    num_spaces = len(spaces)
    if num_query_points is None:
        num_query_points = num_spaces  # Assume one query point per subspace.

    assert num_query_points % num_spaces == 0, (
        f"Number of query points ({num_query_points}) must be a multiple of the number of "
        f"spaces ({num_spaces})."
    )
    num_query_points_per_space = num_query_points // num_spaces

    # If there are local datasets, use them to generate the colors for the query points.
    # Otherwise, use the global dataset and assume the last `num_query_points` points are new.
    if len(history.datasets) > 1:
        # Expect there to be an objective dataset for each subspace.
        datasets = [history.datasets[LocalizedTag(OBJECTIVE, i)] for i in range(num_spaces)]

        _new_points_mask = [
            np.zeros(dataset.query_points.shape[0], dtype=bool) for dataset in datasets
        ]
        # New points in each dataset are at the end.
        for mask in _new_points_mask:
            mask[-num_query_points_per_space:] = True
        # Concatenate the masks.
        new_points_mask = np.concatenate(_new_points_mask)

        if num_init is not None:
            _num_init_mask = [
                np.zeros(dataset.query_points.shape[0], dtype=bool) for dataset in datasets
            ]
            # First num_init points in each dataset are the init points.
            for mask in _num_init_mask:
                mask[:num_init] = True
            # Concatenate the masks.
            num_init = np.concatenate(_num_init_mask)

        # Get the overall query points.
        query_points = np.concatenate([dataset.query_points for dataset in datasets])
    else:
        query_points = history.dataset.query_points  # All query points.
        new_points_mask = np.zeros(query_points.shape[0], dtype=bool)
        new_points_mask[-num_query_points:] = True

    # Plot trust regions.
    space_colors = [rgb2hex(color) for color in cm.rainbow(np.linspace(0, 1, num_spaces))]
    for i, space in enumerate(spaces):
        lb = space.lower
        ub = space.upper
        ax[0, 0].add_patch(
            Rectangle(
                (lb[0], lb[1]),
                ub[0] - lb[0],
                ub[1] - lb[1],
                facecolor=space_colors[i],
                edgecolor=space_colors[i],
                alpha=alpha,
            )
        )

    # Plot new query points, using failure mask to color them.
    point_colors = (
        np.reshape(
            space_colors * num_query_points_per_space, [num_query_points_per_space, num_spaces]
        )
        .flatten("F")
        .tolist()
    )
    plot_bo_points(
        query_points,
        ax[0, 0],
        num_init,
        mask_fail=new_points_mask,
        c_pass="black",
        c_fail=point_colors,
    )

    return fig, ax


def convert_figure_to_frame(fig: plt.Figure) -> TensorType:
    """
    Converts a matplotlib figure to an array of pixels.

    :param fig: a matplotlib figure
    :return: an array of pixels - a frame
    """
    fig.canvas.draw()
    size_pix = fig.get_size_inches() * fig.dpi
    image = np.frombuffer(fig.canvas.tostring_rgb(), dtype="uint8")
    return image.reshape(list(size_pix[::-1].astype(int)) + [3])


def convert_frames_to_gif(frames: Sequence[TensorType], duration: int = 5000) -> io.BytesIO:
    """
    Converts a sequence of frames (arrays of pixels) to a gif.

    :param frames: sequence of frames
    :param duration: duration of each frame in milliseconds
    :return: gif file
    """
    gif_file = io.BytesIO()
    imageio.mimsave(gif_file, frames, format="gif", loop=0, duration=duration)
    return gif_file



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
This package contains the primary interfaces for probabilistic models, :class:`ProbabilisticModel`
and its trainable subclass :class:`TrainableProbabilisticModel`. It also contains tooling for
creating :class:`TrainableProbabilisticModel`\ s from config.
"""

from . import gpflow, gpflux, keras, optimizer
from .interfaces import (
    FastUpdateModel,
    ModelStack,
    ProbabilisticModel,
    ProbabilisticModelType,
    ReparametrizationSampler,
    SupportsCovarianceWithTopFidelity,
    TrainableModelStack,
    TrainableProbabilisticModel,
    TrajectoryFunction,
    TrajectoryFunctionClass,
    TrajectorySampler,
)



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module is the home of the sampling functionality required by Trieste's
GPflow wrappers.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Optional, Tuple, TypeVar, Union, cast

import tensorflow as tf
import tensorflow_probability as tfp
from check_shapes import check_shapes
from gpflow.kernels import Kernel, MultioutputKernel
from gpflux.layers.basis_functions.fourier_features import RandomFourierFeaturesCosine
from gpflux.math import compute_A_inv_b
from typing_extensions import Protocol, TypeGuard, runtime_checkable

from ...space import EncoderFunction
from ...types import TensorType
from ...utils import DEFAULTS, flatten_leading_dims
from ...utils.misc import ensure_positive
from ..interfaces import (
    ProbabilisticModel,
    ReparametrizationSampler,
    SupportsGetInducingVariables,
    SupportsGetInternalData,
    SupportsGetKernel,
    SupportsGetMeanFunction,
    SupportsGetObservationNoise,
    SupportsPredictJoint,
    TrajectoryFunction,
    TrajectoryFunctionClass,
    TrajectorySampler,
    get_encoder,
)

_IntTensorType = Union[tf.Tensor, int]


def qmc_normal_samples(
    num_samples: _IntTensorType,
    n_sample_dim: _IntTensorType,
    skip: _IntTensorType = 0,
    dtype: tf.DType = tf.float64,
) -> tf.Tensor:
    """
    Generates `num_samples` sobol samples, skipping the first `skip`, where each
    sample has dimension `n_sample_dim`.
    """

    if num_samples == 0 or n_sample_dim == 0:
        return tf.zeros(shape=(num_samples, n_sample_dim), dtype=dtype)

    sobol_samples = tf.math.sobol_sample(
        dim=n_sample_dim,
        num_results=num_samples,
        dtype=dtype,
        skip=skip,
    )

    dist = tfp.distributions.Normal(
        loc=tf.constant(0.0, dtype=dtype),
        scale=tf.constant(1.0, dtype=dtype),
    )
    normal_samples = dist.quantile(sobol_samples)
    return normal_samples


class IndependentReparametrizationSampler(ReparametrizationSampler[ProbabilisticModel]):
    r"""
    This sampler employs the *reparameterization trick* to approximate samples from a
    :class:`ProbabilisticModel`\ 's predictive distribution as

    .. math:: x \mapsto \mu(x) + \epsilon \sigma(x)

    where :math:`\epsilon \sim \mathcal N (0, 1)` is constant for a given sampler, thus ensuring
    samples form a continuous curve.
    """

    skip: TensorType = tf.Variable(0, trainable=False)
    """Number of sobol sequence points to skip. This is incremented for each sampler."""

    def __init__(
        self,
        sample_size: int,
        model: ProbabilisticModel,
        qmc: bool = False,
        qmc_skip: bool = True,
        predict_fn: Optional[Callable[[TensorType], tuple[TensorType, TensorType]]] = None,
    ):
        """
        :param sample_size: The number of samples to take at each point. Must be positive.
        :param model: The model to sample from.
        :param qmc: Whether to use QMC sobol sampling instead of random normal sampling. QMC
            sampling more accurately approximates a normal distribution than truly random samples.
        :param qmc_skip: Whether to use the skip parameter to ensure the QMC sampler gives different
            samples whenever it is reset. This is not supported with XLA.
        :param predict_fn: A function to use instead of the model's predict method.
        :raise ValueError (or InvalidArgumentError): If ``sample_size`` is not positive.
        """
        super().__init__(sample_size, model)
        self._eps: Optional[tf.Variable] = None
        self._predict_fn = predict_fn or self._model.predict
        self._qmc = qmc
        self._qmc_skip = qmc_skip

    @check_shapes(
        "at: [N..., 1, D] # IndependentReparametrizationSampler only supports batch sizes of one",
        "return: [N..., S, 1, L]",
    )
    def sample(self, at: TensorType, *, jitter: float = -1.0) -> TensorType:
        """
        Return approximate samples from the `model` specified at :meth:`__init__`. Multiple calls to
        :meth:`sample`, for any given :class:`IndependentReparametrizationSampler` and ``at``, will
        produce the exact same samples. Calls to :meth:`sample` on *different*
        :class:`IndependentReparametrizationSampler` instances will produce different samples.

        :param at: Where to sample the predictive distribution, with shape `[..., 1, D]`, for points
            of dimension `D`.
        :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
            the covariance matrix. If a negative value is passed then no jitter is applied but
            all values are capped to a hardcoded minimum.
        :return: The samples, of shape `[..., S, 1, L]`, where `S` is the `sample_size` and `L` is
            the number of latent model dimensions.
        :raise ValueError (or InvalidArgumentError): If ``at`` has an invalid shape or ``jitter``
            is negative.
        """
        mean, var = self._predict_fn(at[..., None, :, :])  # [..., 1, 1, L], [..., 1, 1, L]
        var = ensure_positive(var) if jitter < 0 else (var + jitter)

        def sample_eps() -> tf.Tensor:
            self._initialized.assign(True)
            if self._qmc:
                if self._qmc_skip:
                    skip = IndependentReparametrizationSampler.skip
                    IndependentReparametrizationSampler.skip.assign(skip + self._sample_size)
                else:
                    skip = tf.constant(0)
                normal_samples = qmc_normal_samples(
                    self._sample_size, mean.shape[-1], skip, dtype=var.dtype
                )
            else:
                normal_samples = tf.random.normal(
                    [self._sample_size, tf.shape(mean)[-1]], dtype=var.dtype
                )
            return normal_samples  # [S, L]

        if self._eps is None:
            self._eps = tf.Variable(sample_eps())

        tf.cond(
            self._initialized,
            lambda: self._eps,
            lambda: self._eps.assign(sample_eps()),
        )

        return mean + tf.sqrt(var) * self._eps[:, None, :]  # [..., S, 1, L]


class BatchReparametrizationSampler(ReparametrizationSampler[SupportsPredictJoint]):
    r"""
    This sampler employs the *reparameterization trick* to approximate batches of samples from a
    :class:`ProbabilisticModel`\ 's predictive joint distribution as

    .. math:: x \mapsto \mu(x) + \epsilon L(x)

    where :math:`L` is the Cholesky factor s.t. :math:`LL^T` is the covariance, and
    :math:`\epsilon \sim \mathcal N (0, 1)` is constant for a given sampler, thus ensuring samples
    form a continuous curve.
    """

    skip: TensorType = tf.Variable(0, trainable=False)
    """Number of sobol sequence points to skip. This is incremented for each sampler."""

    def __init__(
        self,
        sample_size: int,
        model: SupportsPredictJoint,
        qmc: bool = False,
        qmc_skip: bool = True,
    ):
        """
        :param sample_size: The number of samples for each batch of points. Must be positive.
        :param model: The model to sample from.
        :param qmc: Whether to use QMC sobol sampling instead of random normal sampling. QMC
            sampling more accurately approximates a normal distribution than truly random samples.
        :param qmc_skip: Whether to use the skip parameter to ensure the QMC sampler gives different
            samples whenever it is reset. This is not supported with XLA.
        :raise ValueError (or InvalidArgumentError): If ``sample_size`` is not positive.
        """
        super().__init__(sample_size, model)
        if not isinstance(model, SupportsPredictJoint):
            raise NotImplementedError(
                f"BatchReparametrizationSampler only works with models that support "
                f"predict_joint; received {model!r}"
            )
        self._eps: Optional[tf.Variable] = None
        self._qmc = qmc
        self._qmc_skip = qmc_skip

    def sample(self, at: TensorType, *, jitter: float = DEFAULTS.JITTER) -> TensorType:
        """
        Return approximate samples from the `model` specified at :meth:`__init__`. Multiple calls to
        :meth:`sample`, for any given :class:`BatchReparametrizationSampler` and ``at``, will
        produce the exact same samples. Calls to :meth:`sample` on *different*
        :class:`BatchReparametrizationSampler` instances will produce different samples.

        :param at: Batches of query points at which to sample the predictive distribution, with
            shape `[..., B, D]`, for batches of size `B` of points of dimension `D`. Must have a
            consistent batch size across all calls to :meth:`sample` for any given
            :class:`BatchReparametrizationSampler`.
        :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
            the covariance matrix.
        :return: The samples, of shape `[..., S, B, L]`, where `S` is the `sample_size`, `B` the
            number of points per batch, and `L` the dimension of the model's predictive
            distribution.
        :raise ValueError (or InvalidArgumentError): If any of the following are true:
            - ``at`` is a scalar.
            - The batch size `B` of ``at`` is not positive.
            - The batch size `B` of ``at`` differs from that of previous calls.
            - ``jitter`` is negative.
        """
        tf.debugging.assert_rank_at_least(at, 2)
        tf.debugging.assert_greater_equal(jitter, 0.0)

        batch_size = at.shape[-2]

        tf.debugging.assert_positive(batch_size)

        mean, cov = self._model.predict_joint(at)  # [..., B, L], [..., L, B, B]

        def sample_eps() -> tf.Tensor:
            self._initialized.assign(True)
            if self._qmc:
                if self._qmc_skip:
                    skip = IndependentReparametrizationSampler.skip
                    IndependentReparametrizationSampler.skip.assign(skip + self._sample_size)
                else:
                    skip = tf.constant(0)
                normal_samples = qmc_normal_samples(
                    self._sample_size * mean.shape[-1], batch_size, skip, dtype=cov.dtype
                )  # [S*L, B]
                normal_samples = tf.reshape(
                    normal_samples, (mean.shape[-1], self._sample_size, batch_size)
                )  # [L, S, B]
                normal_samples = tf.transpose(normal_samples, perm=[0, 2, 1])  # [L, B, S]
            else:
                normal_samples = tf.random.normal(
                    [tf.shape(mean)[-1], batch_size, self._sample_size], dtype=cov.dtype
                )  # [L, B, S]
            return normal_samples

        if self._eps is None:
            # dynamically shaped as the same sampler may be called with different sized batches
            self._eps = tf.Variable(sample_eps(), shape=[None, None, self._sample_size])

        tf.cond(
            self._initialized,
            lambda: self._eps,
            lambda: self._eps.assign(sample_eps()),
        )

        if self._initialized:
            tf.debugging.assert_equal(
                batch_size,
                tf.shape(self._eps)[-2],
                f"{type(self).__name__} requires a fixed batch size. Got batch size {batch_size}"
                f" but previous batch size was {tf.shape(self._eps)[-2]}.",
            )

        identity = tf.eye(batch_size, dtype=cov.dtype)  # [B, B]
        cov_cholesky = tf.linalg.cholesky(cov + jitter * identity)  # [..., L, B, B]

        variance_contribution = cov_cholesky @ self._eps  # [..., L, B, S]

        leading_indices = tf.range(tf.rank(variance_contribution) - 3)
        absolute_trailing_indices = [-1, -2, -3] + tf.rank(variance_contribution)
        new_order = tf.concat([leading_indices, absolute_trailing_indices], axis=0)

        return mean[..., None, :, :] + tf.transpose(variance_contribution, new_order)


@runtime_checkable
class FeatureDecompositionInternalDataModel(
    SupportsGetKernel,
    SupportsGetMeanFunction,
    SupportsGetObservationNoise,
    SupportsGetInternalData,
    Protocol,
):
    """
    A probabilistic model that supports get_kernel, get_mean_function, get_observation_noise
    and get_internal_data methods.
    """


@runtime_checkable
class FeatureDecompositionInducingPointModel(
    SupportsGetKernel, SupportsGetMeanFunction, SupportsGetInducingVariables, Protocol
):
    """
    A probabilistic model that supports get_kernel, get_mean_function
    and get_inducing_point methods.
    """


FeatureDecompositionTrajectorySamplerModel = Union[
    FeatureDecompositionInducingPointModel,
    FeatureDecompositionInternalDataModel,
]

FeatureDecompositionTrajectorySamplerModelType = TypeVar(
    "FeatureDecompositionTrajectorySamplerModelType",
    bound=FeatureDecompositionTrajectorySamplerModel,
    contravariant=True,
)


def _is_multioutput_kernel(kernel: Kernel) -> TypeGuard[MultioutputKernel]:
    return isinstance(kernel, MultioutputKernel)


def _get_kernel_function(kernel: Kernel) -> Callable[[TensorType, TensorType], tf.Tensor]:
    # Select between a multioutput kernel and a single-output kernel.
    def K(X: TensorType, X2: Optional[TensorType] = None) -> tf.Tensor:
        if _is_multioutput_kernel(kernel):
            return kernel(X, X2, full_cov=True, full_output_cov=False)  # [L, M, M]
        else:
            return tf.expand_dims(kernel(X, X2), axis=0)  # [1, M, M]

    return K


class FeatureDecompositionTrajectorySampler(
    TrajectorySampler[FeatureDecompositionTrajectorySamplerModelType],
    ABC,
):
    r"""

    This is a general class to build functions that approximate a trajectory sampled from an
    underlying Gaussian process model.

    In particular, we approximate the Gaussian processes' posterior samples as the finite feature
    approximation

    .. math:: \hat{f}(x) = \sum_{i=1}^m \phi_i(x)\theta_i

    where :math:`\phi_i` are m features and :math:`\theta_i` are feature weights sampled from a
    given distribution

    Achieving consistency (ensuring that the same sample draw for all evalutions of a particular
    trajectory function) for exact sample draws from a GP is prohibitively costly because it scales
    cubically with the number of query points. However, finite feature representations can be
    evaluated with constant cost regardless of the required number of queries.
    """

    def __init__(
        self,
        model: FeatureDecompositionTrajectorySamplerModelType,
        feature_functions: ResampleableRandomFourierFeatureFunctions,
    ):
        """
        :param model: The model to sample from.
        :raise ValueError: If ``dataset`` is empty.
        """

        super().__init__(model)
        self._feature_functions = feature_functions
        self._weight_sampler: Optional[Callable[[int], TensorType]] = None  # lazy init
        self._mean_function = model.get_mean_function()

    def __repr__(self) -> str:
        """"""
        return f"""{self.__class__.__name__}(
        {self._model!r},
        {self._feature_functions!r})
        """

    def get_trajectory(self) -> TrajectoryFunction:
        """
        Generate an approximate function draw (trajectory) by sampling weights
        and evaluating the feature functions.

        :return: A trajectory function representing an approximate trajectory from the Gaussian
            process, taking an input of shape `[N, B, D]` and returning shape `[N, B, L]`
            where `L` is the number of outputs of the model.
        """

        weight_sampler = self._prepare_weight_sampler()  # prep feature weight distribution

        return feature_decomposition_trajectory(
            feature_functions=self._feature_functions,
            weight_sampler=weight_sampler,
            mean_function=self._mean_function,
            encoder=get_encoder(self._model),
        )

    def update_trajectory(self, trajectory: TrajectoryFunction) -> TrajectoryFunction:
        """
        Efficiently update a :const:`TrajectoryFunction` to reflect an update in its
        underlying :class:`ProbabilisticModel` and resample accordingly.

        For a :class:`FeatureDecompositionTrajectorySampler`, updating the sampler
        corresponds to resampling the feature functions (taking into account any
        changed kernel parameters) and recalculating the weight distribution.

        :param trajectory: The trajectory function to be resampled.
        :return: The new resampled trajectory function.
        """
        tf.debugging.Assert(
            isinstance(trajectory, feature_decomposition_trajectory), [tf.constant([])]
        )

        self._feature_functions.resample()  # resample Fourier feature decomposition
        weight_sampler = self._prepare_weight_sampler()  # recalculate weight distribution

        cast(feature_decomposition_trajectory, trajectory).update(weight_sampler=weight_sampler)

        return trajectory  # return trajectory with updated features and weight distribution

    def resample_trajectory(self, trajectory: TrajectoryFunction) -> TrajectoryFunction:
        """
        Efficiently resample a :const:`TrajectoryFunction` in-place to avoid function retracing
        with every new sample.

        :param trajectory: The trajectory function to be resampled.
        :return: The new resampled trajectory function.
        """
        tf.debugging.Assert(
            isinstance(trajectory, feature_decomposition_trajectory), [tf.constant([])]
        )
        cast(feature_decomposition_trajectory, trajectory).resample()
        return trajectory  # return trajectory with resampled weights

    @abstractmethod
    def _prepare_weight_sampler(self) -> Callable[[int], TensorType]:  # [B] -> [B, F, L]
        """
        Calculate the posterior of the feature weights for the specified feature functions,
        returning a function that takes in a batch size `B` and returns `B` samples for
        the weights of each of the `F` features for `L` outputs.
        """
        raise NotImplementedError


class RandomFourierFeatureTrajectorySampler(
    FeatureDecompositionTrajectorySampler[FeatureDecompositionInternalDataModel]
):
    r"""
    This class builds functions that approximate a trajectory sampled from an underlying Gaussian
    process model. For tractibility, the Gaussian process is approximated with a Bayesian
    Linear model across a set of features sampled from the Fourier feature decomposition of
    the model's kernel. See :cite:`hernandez2014predictive` for details. Currently we do not
    support models with multiple latent Gaussian processes.

    In particular, we approximate the Gaussian processes' posterior samples as the finite feature
    approximation

    .. math:: \hat{f}(x) = \sum_{i=1}^m \phi_i(x)\theta_i

    where :math:`\phi_i` are m Fourier features and :math:`\theta_i` are
    feature weights sampled from a posterior distribution that depends on the feature values at the
    model's datapoints.

    Our implementation follows :cite:`hernandez2014predictive`, with our calculations
    differing slightly depending on properties of the problem. In particular,  we used different
    calculation strategies depending on the number of considered features m and the number
    of data points n.

    If :math:`m<n` then we follow Appendix A of :cite:`hernandez2014predictive` and calculate the
    posterior distribution for :math:`\theta` following their Bayesian linear regression motivation,
    i.e. the computation revolves around an O(m^3)  inversion of a design matrix.

    If :math:`n<m` then we use the kernel trick to recast computation to revolve around an O(n^3)
    inversion of a gram matrix. As well as being more efficient in early BO
    steps (where :math:`n<m`), this second computation method allows much larger choices
    of m (as required to approximate very flexible kernels).
    """

    def __init__(
        self,
        model: FeatureDecompositionInternalDataModel,
        num_features: int = 1000,
    ):
        """
        :param model: The model to sample from.
        :param num_features: The number of features used to approximate the kernel. We use a default
            of 1000 as it typically perfoms well for a wide range of kernels. Note that very smooth
            kernels (e.g. RBF) can be well-approximated with fewer features.
        :raise ValueError: If ``dataset`` is empty.
        """

        if not isinstance(model, FeatureDecompositionInternalDataModel):
            raise NotImplementedError(
                f"RandomFourierFeatureTrajectorySampler only works with models with "
                f"get_kernel, get_observation_noise and get_internal_data methods; "
                f"but received {model!r}."
            )

        tf.debugging.assert_positive(num_features)
        self._num_features = num_features
        feature_functions = ResampleableRandomFourierFeatureFunctions(model, self._num_features)
        super().__init__(model, feature_functions)

    def _prepare_weight_sampler(self) -> Callable[[int], TensorType]:  # [B] -> [B, F, 1]
        """
        Calculate the posterior of theta (the feature weights) for the RFFs, returning
        a function that takes in a batch size `B` and returns `B` samples for
        the weights of each of the RFF `F` features for one output.
        """

        dataset = self._model.get_internal_data()
        num_data = tf.shape(dataset.query_points)[0]  # n
        if (
            self._num_features < num_data
        ):  # if m < n  then calculate posterior in design space (an m*m matrix inversion)
            theta_posterior = self._prepare_theta_posterior_in_design_space()
        else:  # if n <= m  then calculate posterior in gram space (an n*n matrix inversion)
            theta_posterior = self._prepare_theta_posterior_in_gram_space()

        return lambda b: tf.expand_dims(theta_posterior.sample(b), axis=-1)

    def _prepare_theta_posterior_in_design_space(self) -> tfp.distributions.MultivariateNormalTriL:
        r"""
        Calculate the posterior of theta (the feature weights) in the design space. This
        distribution is a Gaussian

        .. math:: \theta \sim N(D^{-1}\Phi^Ty,D^{-1}\sigma^2)

        where the [m,m] design matrix :math:`D=(\Phi^T\Phi + \sigma^2I_m)` is defined for
        the [n,m] matrix of feature evaluations across the training data :math:`\Phi`
        and observation noise variance :math:`\sigma^2`.
        """
        dataset = self._model.get_internal_data()
        phi = self._feature_functions(tf.convert_to_tensor(dataset.query_points))  # [n, m]
        D = tf.matmul(phi, phi, transpose_a=True)  # [m, m]
        s = self._model.get_observation_noise() * tf.eye(self._num_features, dtype=phi.dtype)
        L = tf.linalg.cholesky(D + s)
        D_inv = tf.linalg.cholesky_solve(L, tf.eye(self._num_features, dtype=phi.dtype))

        residuals = dataset.observations - self._model.get_mean_function()(dataset.query_points)
        theta_posterior_mean = tf.matmul(D_inv, tf.matmul(phi, residuals, transpose_a=True))[
            :, 0
        ]  # [m,]
        theta_posterior_chol_covariance = tf.linalg.cholesky(
            D_inv * self._model.get_observation_noise()
        )  # [m, m]

        return tfp.distributions.MultivariateNormalTriL(
            theta_posterior_mean, theta_posterior_chol_covariance
        )

    def _prepare_theta_posterior_in_gram_space(self) -> tfp.distributions.MultivariateNormalTriL:
        r"""
        Calculate the posterior of theta (the feature weights) in the gram space.

         .. math:: \theta \sim N(\Phi^TG^{-1}y,I_m - \Phi^TG^{-1}\Phi)

        where the [n,n] gram matrix :math:`G=(\Phi\Phi^T + \sigma^2I_n)` is defined for the [n,m]
        matrix of feature evaluations across the training data :math:`\Phi` and
        observation noise variance :math:`\sigma^2`.
        """
        dataset = self._model.get_internal_data()
        num_data = tf.shape(dataset.query_points)[0]  # n
        phi = self._feature_functions(tf.convert_to_tensor(dataset.query_points))  # [n, m]
        G = tf.matmul(phi, phi, transpose_b=True)  # [n, n]
        s = self._model.get_observation_noise() * tf.eye(num_data, dtype=phi.dtype)
        L = tf.linalg.cholesky(G + s)
        L_inv_phi = tf.linalg.triangular_solve(L, phi)  # [n, m]
        residuals = dataset.observations - self._model.get_mean_function()(
            dataset.query_points
        )  # [n, 1]
        L_inv_y = tf.linalg.triangular_solve(L, residuals)  # [n, 1]

        theta_posterior_mean = tf.tensordot(tf.transpose(L_inv_phi), L_inv_y, [[-1], [-2]])[
            :, 0
        ]  # [m,]
        theta_posterior_covariance = tf.eye(self._num_features, dtype=phi.dtype) - tf.tensordot(
            tf.transpose(L_inv_phi), L_inv_phi, [[-1], [-2]]
        )  # [m, m]
        theta_posterior_chol_covariance = tf.linalg.cholesky(theta_posterior_covariance)  # [m, m]

        return tfp.distributions.MultivariateNormalTriL(
            theta_posterior_mean, theta_posterior_chol_covariance
        )


class DecoupledTrajectorySampler(
    FeatureDecompositionTrajectorySampler[
        Union[
            FeatureDecompositionInducingPointModel,
            FeatureDecompositionInternalDataModel,
        ]
    ]
):
    r"""

    This class builds functions that approximate a trajectory sampled from an underlying Gaussian
    process model using decoupled sampling. See :cite:`wilson2020efficiently` for an introduction
    to decoupled sampling.

    Unlike our :class:`RandomFourierFeatureTrajectorySampler` which uses a RFF decomposition to
    aprroximate the Gaussian process posterior, a :class:`DecoupledTrajectorySampler` only
    uses an RFF decomposition to approximate the Gausian process prior and instead using
    a canonical decomposition to discretize the effect of updating the prior on the given data.

    In particular, we approximate the Gaussian processes' posterior samples as the finite feature
    approximation

    .. math:: \hat{f}(.) = \sum_{i=1}^L w_i\phi_i(.) + \sum_{j=1}^m v_jk(.,z_j)

    where :math:`\phi_i(.)` and :math:`w_i` are the Fourier features and their weights that
    discretize the prior. In contrast, `k(.,z_j)` and :math:`v_i` are the canonical features and
    their weights that discretize the data update.

    The expression for :math:`v_i` depends on if we are using an exact Gaussian process or a sparse
    approximations. See  eq. (13) in :cite:`wilson2020efficiently` for details.

    Note that if a model is both of :class:`FeatureDecompositionInducingPointModel` type and
    :class:`FeatureDecompositionInternalDataModel` type,
    :class:`FeatureDecompositionInducingPointModel` will take a priority and inducing points
    will be used for computations rather than data.
    """

    def __init__(
        self,
        model: Union[
            FeatureDecompositionInducingPointModel,
            FeatureDecompositionInternalDataModel,
        ],
        num_features: int = 1000,
    ):
        """
        :param model: The model to sample from.
        :param num_features: The number of features used to approximate the kernel. We use a default
            of 1000 as it typically perfoms well for a wide range of kernels. Note that very smooth
            kernels (e.g. RBF) can be well-approximated with fewer features.
        :raise NotImplementedError: If the model is not of valid type.
        """
        if not isinstance(
            model, (FeatureDecompositionInducingPointModel, FeatureDecompositionInternalDataModel)
        ):
            raise NotImplementedError(
                f"DecoupledTrajectorySampler only works with models that either support "
                f"get_kernel, get_observation_noise and get_internal_data or support get_kernel "
                f"and get_inducing_variables; but received {model!r}."
            )

        tf.debugging.assert_positive(num_features)
        self._num_features = num_features
        feature_functions = ResampleableDecoupledFeatureFunctions(model, self._num_features)

        super().__init__(model, feature_functions)

    def _prepare_weight_sampler(self) -> Callable[[int], TensorType]:  # [B] -> [B, F + M, L]
        """
        Prepare the sampler function that provides samples of the feature weights
        for both the RFF and canonical feature functions, i.e. we return a function
        that takes in a batch size `B` and returns `B` samples for the weights of each of
        the `F`  RFF features and `M` canonical features for `L` outputs.
        """

        kernel_K = _get_kernel_function(self._model.get_kernel())
        if isinstance(self._model, FeatureDecompositionInducingPointModel):
            (  # extract variational parameters
                inducing_points,
                q_mu,
                q_sqrt,
                whiten,
            ) = self._model.get_inducing_variables()  # [M, D], [M, L], [L, M, M], []
            Kmm = kernel_K(inducing_points, inducing_points)  # [L, M, M]
            Kmm += tf.eye(tf.shape(inducing_points)[0], dtype=Kmm.dtype) * DEFAULTS.JITTER
        else:  # massage quantities from GP to look like variational parameters
            internal_data = self._model.get_internal_data()
            inducing_points = internal_data.query_points  # [M, D]
            q_mu = self._model.get_internal_data().observations  # [M, L]
            q_mu = q_mu - self._model.get_mean_function()(
                inducing_points
            )  # account for mean function
            q_sqrt = tf.eye(tf.shape(inducing_points)[0], dtype=tf.float64)  # [M, M]
            q_sqrt = tf.expand_dims(q_sqrt, axis=0)  # [1, M, M]
            q_sqrt = tf.math.sqrt(self._model.get_observation_noise()) * q_sqrt
            whiten = False
            Kmm = kernel_K(inducing_points, inducing_points) + q_sqrt**2  # [L, M, M]

        M, L = tf.shape(q_mu)
        tf.debugging.assert_shapes(
            [
                (inducing_points, ["M", "D"]),
                (q_mu, ["M", "L"]),
                (q_sqrt, ["L", "M", "M"]),
                (Kmm, ["L", "M", "M"]),
            ]
        )

        def weight_sampler(batch_size: int) -> Tuple[TensorType, TensorType]:
            prior_weights = tf.random.normal(  # Non-RFF features will require scaling here
                [L, self._num_features, batch_size], dtype=tf.float64
            )  # [L, F, B]

            u_noise_sample = tf.matmul(
                q_sqrt,  # [L, M, M]
                tf.random.normal((L, M, batch_size), dtype=tf.float64),  # [L, M, B]
            )  # [L, M, B]

            u_sample = tf.linalg.matrix_transpose(q_mu)[..., None] + u_noise_sample  # [L, M, B]

            if whiten:
                Luu = tf.linalg.cholesky(Kmm)  # [L, M, M]
                u_sample = tf.matmul(Luu, u_sample)  # [L, M, B]

            # It is important that the feature-function is called with a tensor, instead of a
            # parameter (which inducing points can be). This is to ensure pickling works correctly.
            # First time a Keras layer (i.e. feature-functions) is built, the shape of the input is
            # used to set the input-spec. If the input is a parameter, the input-spec will not be
            # for an ordinary tensor and pickling will fail.
            phi_Z = self._feature_functions(tf.convert_to_tensor(inducing_points))[
                ..., : self._num_features
            ]  # [M, F] or [L, M, F]
            weight_space_prior_Z = phi_Z @ prior_weights  # [L, M, B]

            diff = u_sample - weight_space_prior_Z  # [L, M, B]

            v = compute_A_inv_b(Kmm, diff)  # [L, M, B]

            tf.debugging.assert_shapes([(v, ["L", "M", "B"]), (prior_weights, ["L", "F", "B"])])

            return tf.transpose(
                tf.concat([prior_weights, v], axis=1), perm=[2, 1, 0]
            )  # [B, F + M, L]

        return weight_sampler


class ResampleableRandomFourierFeatureFunctions(RandomFourierFeaturesCosine):
    """
    A wrapper around GPFlux's random Fourier feature function that allows for
    efficient in-place updating when generating new decompositions.

    In particular, the bias and weights are stored as variables, which can then be
    updated by calling :meth:`resample` without triggering expensive graph retracing.

    Note that if a model is both of :class:`FeatureDecompositionInducingPointModel` type and
    :class:`FeatureDecompositionInternalDataModel` type,
    :class:`FeatureDecompositionInducingPointModel` will take a priority and inducing points
    will be used for computations rather than data.
    """

    def __init__(
        self,
        model: Union[
            FeatureDecompositionInducingPointModel,
            FeatureDecompositionInternalDataModel,
        ],
        n_components: int,
    ):
        """
        :param model: The model that will be approximed by these feature functions.
        :param n_components: The desired number of features.
        :raise NotImplementedError: If the model is not of valid type.
        """
        if not isinstance(
            model,
            (
                FeatureDecompositionInducingPointModel,
                FeatureDecompositionInternalDataModel,
            ),
        ):
            raise NotImplementedError(
                f"ResampleableRandomFourierFeatureFunctions only work with models that either"
                f"support get_kernel, get_observation_noise and get_internal_data or support "
                f"get_kernel and get_inducing_variables;"
                f"but received {model!r}."
            )

        super().__init__(model.get_kernel(), n_components, dtype=tf.float64)

        if isinstance(model, SupportsGetInducingVariables):
            dummy_X = model.get_inducing_variables()[0][0:1, :]
        else:
            dummy_X = model.get_internal_data().query_points[0:1, :]
        dummy_X = self.kernel.slice(dummy_X, None)[0]  # Keep only the active dims from the kernel.

        # Always build the weights and biases. This is important for saving the trajectory (using
        # tf.saved_model.save) before it has been used.
        self.build(dummy_X.shape)

    def resample(self) -> None:
        """
        Resample weights and biases
        """
        self.b.assign(self._bias_init(tf.shape(self.b), dtype=self._dtype))
        self.W.assign(self._weights_init(tf.shape(self.W), dtype=self._dtype))

    def call(self, inputs: TensorType) -> TensorType:  # [N, D] -> [N, F] or [L, N, F]
        """
        Evaluate the basis functions at ``inputs``
        """
        inputs = self.kernel.slice(inputs, None)[0]  # Keep only active dims from the kernel
        return super().call(inputs)  # [N, F] or [L, N, F]


class ResampleableDecoupledFeatureFunctions(ResampleableRandomFourierFeatureFunctions):
    """
    A wrapper around our :class:`ResampleableRandomFourierFeatureFunctions` which rather
    than evaluates just `F` RFF functions instead evaluates the concatenation of
    `F` RFF functions with evaluations of the canonical basis functions.

    Note that if a model is both of :class:`FeatureDecompositionInducingPointModel` type and
    :class:`FeatureDecompositionInternalDataModel` type,
    :class:`FeatureDecompositionInducingPointModel` will take a priority and inducing points
    will be used for computations rather than data.
    """

    def __init__(
        self,
        model: Union[
            FeatureDecompositionInducingPointModel,
            FeatureDecompositionInternalDataModel,
        ],
        n_components: int,
    ):
        """
        :param model: The model that will be approximed by these feature functions.
        :param n_components: The desired number of features.
        """

        super().__init__(model, n_components)

        if isinstance(model, SupportsGetInducingVariables):
            self._inducing_points = model.get_inducing_variables()[0]  # [M, D]
        else:
            self._inducing_points = model.get_internal_data().query_points  # [M, D]

        kernel_K = _get_kernel_function(self.kernel)
        self._canonical_feature_functions = lambda x: tf.linalg.matrix_transpose(
            kernel_K(self._inducing_points, x)
        )

    def call(self, inputs: TensorType) -> TensorType:  # [N, D] -> [N, F + M] or [L, N, F + M]
        """
        combine prior basis functions with canonical basis functions
        """
        fourier_feature_eval = super().call(inputs)  # [N, F] or [L, N, F]
        canonical_feature_eval = self._canonical_feature_functions(inputs)  # [1, N, M] or [L, N, M]
        # ensure matching rank between features, i.e. drop the leading 1 dimension
        matched_shape = tf.shape(canonical_feature_eval)[-tf.rank(fourier_feature_eval) :]
        canonical_feature_eval = tf.reshape(canonical_feature_eval, matched_shape)
        return tf.concat([fourier_feature_eval, canonical_feature_eval], axis=-1)


class feature_decomposition_trajectory(TrajectoryFunctionClass):
    r"""
    An approximate sample from a Gaussian processes' posterior samples represented as a
    finite weighted sum of features.

    A trajectory is given by

    .. math:: \hat{f}(x) = \sum_{i=1}^m \phi_i(x)\theta_i

    where :math:`\phi_i` are m feature functions and :math:`\theta_i` are
    feature weights sampled from a posterior distribution.

    The number of trajectories (i.e. batch size) is determined from the first call of the
    trajectory. In order to change the batch size, a new :class:`TrajectoryFunction` must be built.
    """

    def __init__(
        self,
        feature_functions: Callable[[TensorType], TensorType],
        weight_sampler: Callable[[int], TensorType],
        mean_function: Callable[[TensorType], TensorType],
        encoder: EncoderFunction | None = None,
    ):
        """
        :param feature_functions: Set of feature function.
        :param weight_sampler: New sampler that generates feature weight samples.
        :param mean_function: The underlying model's mean function.
        :param encoder: Optional encoder with which to transform input points.
        """
        self._feature_functions = feature_functions
        self._mean_function = mean_function
        self._weight_sampler = weight_sampler
        self._encoder = encoder
        self._initialized = tf.Variable(False)

        self._weights_sample = tf.Variable(  # dummy init to be updated before trajectory evaluation
            tf.ones([0, 0, 0], dtype=tf.float64), shape=[None, None, None]
        )

        self._batch_size = tf.Variable(
            0, dtype=tf.int32
        )  # dummy init to be updated before trajectory evaluation

    @tf.function
    def __call__(self, inputs: TensorType) -> TensorType:  # [N, B, D] -> [N, B, L]
        """Call trajectory function."""

        if self._encoder is not None:
            inputs = self._encoder(inputs)

        if not self._initialized:  # work out desired batch size from input
            self._batch_size.assign(tf.shape(inputs)[-2])  # B
            self.resample()  # sample B feature weights
            self._initialized.assign(True)

        tf.debugging.assert_equal(
            tf.shape(inputs)[-2],
            self._batch_size.value(),
            message=f"""
            This trajectory only supports batch sizes of {self._batch_size}.
            If you wish to change the batch size you must get a new trajectory
            by calling the get_trajectory method of the trajectory sampler.
            """,
        )

        flat_inputs, unflatten = flatten_leading_dims(inputs)  # [N*B, D]
        flattened_feature_evaluations = self._feature_functions(
            flat_inputs
        )  # [N*B, F + M] or [L, N*B, F + M]
        # ensure tensor is always rank 3
        rank3_shape = tf.concat([[1], tf.shape(flattened_feature_evaluations)], axis=0)[-3:]
        flattened_feature_evaluations = tf.reshape(flattened_feature_evaluations, rank3_shape)
        flattened_feature_evaluations = tf.transpose(
            flattened_feature_evaluations, perm=[1, 2, 0]
        )  # [N*B, F + M, L]
        feature_evaluations = unflatten(flattened_feature_evaluations)  # [N, B, F + M, L]

        mean = self._mean_function(inputs)  # account for the model's mean function
        return tf.reduce_sum(feature_evaluations * self._weights_sample, -2) + mean  # [N, B, L]

    def resample(self) -> None:
        """
        Efficiently resample in-place without retracing.
        """
        self._weights_sample.assign(  # [B, F + M, L]
            self._weight_sampler(self._batch_size)
        )  # resample weights

    def update(self, weight_sampler: Callable[[int], TensorType]) -> None:
        """
        Efficiently update the trajectory with a new weight distribution and resample its weights.

        :param weight_sampler: New sampler that generates feature weight samples.
        """
        self._weight_sampler = weight_sampler  # update weight sampler
        self.resample()  # resample weights
